# -*- coding: utf-8 -*-
# Copyright 2020 Matt Martz

import contextlib
import http.client
import socket
import urllib.error
import urllib.request
import urllib.response

from .errors import HTTPError
from .response import Response


class RedirectHandler(urllib.request.HTTPRedirectHandler):
    def __init__(self, allow_redirects=True):
        self.allow_redirects = allow_redirects

    def redirect_request(self, req, fp, code, msg, headers, newurl):
        if not self.allow_redirects:
            raise HTTPError(newurl, code, msg, headers, fp, req)

        m = req.get_method()
        if code not in (301, 302, 303, 307, 308):
            raise HTTPError(req.full_url, code, msg, headers, fp, req)

        data = req.data
        origin_req_host = req.origin_req_host

        # Be conciliant with URIs containing a space.  This is mainly
        # redundant with the more complete encoding done in http_error_302,
        # but it is kept for compatibility with other callers.
        newurl = newurl.replace(' ', '%20')

        # Suport redirect with payload and original headers
        if code in (307, 308):
            # Preserve payload and headers
            newheaders = req.headers
        else:
            # Do not preserve payload and filter headers
            data = None

            CONTENT_HEADERS = ('content-length', 'content-type')
            newheaders = {k: v for k, v in req.headers.items()
                          if k.lower() not in CONTENT_HEADERS}

            # http://tools.ietf.org/html/rfc7231#section-6.4.4
            if code == 303 and m != 'HEAD':
                m = 'GET'

            # Non-standard transformations

            # Turn 302s into GETs.
            if code == 302 and m != 'HEAD':
                m = 'GET'

            # Turn 301 from POST into GET
            if code == 301 and m == 'POST':
                m = 'GET'

        return urllib.request.Request(
            newurl,
            method=m,
            headers=newheaders,
            data=data,
            origin_req_host=origin_req_host,
            unverifiable=True
        )

    def http_error_302(self, req, fp, code, msg, headers):
        try:
            return super(RedirectHandler, self).http_error_302(
                req,
                fp,
                code,
                msg,
                headers
            )
        except HTTPError:
            urllib.response.addinfourl(fp, headers, req.full_url, code=code)
            return Response(
                fp,
                request=req
            )

    http_error_301 = http_error_303 = http_error_302
    http_error_307 = http_error_308 = http_error_302


class HTTPErrorHandler(urllib.request.HTTPDefaultErrorHandler):
    def http_error_default(self, req, fp, code, msg, hdrs):
        raise HTTPError(req.full_url, code, msg, hdrs, fp, req)


class HTTPSClientAuthHandler(urllib.request.HTTPSHandler):
    '''Handles client authentication via cert/key

    This is a fairly lightweight extension on HTTPSHandler, and can be used
    in place of HTTPSHandler
    '''

    def __init__(self, client_cert=None, client_key=None, unix_socket=None,
                 **kwargs):
        urllib.request.HTTPSHandler.__init__(self, **kwargs)
        self.client_cert = client_cert
        self.client_key = client_key
        self._unix_socket = unix_socket

    def https_open(self, req):
        return self.do_open(self._build_https_connection, req)

    def _build_https_connection(self, host, **kwargs):
        kwargs.update({
            'cert_file': self.client_cert,
            'key_file': self.client_key,
            'context': self._context,
        })
        if self._unix_socket:
            return UnixHTTPSConnection(self._unix_socket)(host, **kwargs)
        return http.client.HTTPSConnection(host, **kwargs)


@contextlib.contextmanager
def unix_socket_patch_httpconnection_connect():
    '''Monkey patch ``http.client.HTTPConnection.connect`` to be
    ``UnixHTTPConnection.connect`` so that when calling
    ``super(UnixHTTPSConnection, self).connect()`` we get the
    correct behavior of creating self.sock for the unix socket
    '''
    _connect = http.client.HTTPConnection.connect
    http.client.HTTPConnection.connect = UnixHTTPConnection.connect
    yield
    http.client.HTTPConnection.connect = _connect


class UnixHTTPSConnection(http.client.HTTPSConnection):
    def __init__(self, unix_socket):
        self._unix_socket = unix_socket

    def connect(self):
        # This method exists simply to ensure we monkeypatch
        # http.client.HTTPConnection.connect to call UnixHTTPConnection.connect
        with unix_socket_patch_httpconnection_connect():
            super(UnixHTTPSConnection, self).connect()

    def __call__(self, *args, **kwargs):
        http.client.HTTPSConnection.__init__(self, *args, **kwargs)
        return self


class UnixHTTPConnection(http.client.HTTPConnection):
    '''Handles http requests to a unix socket file'''

    def __init__(self, unix_socket):
        self._unix_socket = unix_socket

    def connect(self):
        self.sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        try:
            self.sock.connect(self._unix_socket)
        except OSError as e:
            raise OSError(
                'Invalid Socket File (%s): %s' % (self._unix_socket, e)
            )
        if self.timeout is not socket._GLOBAL_DEFAULT_TIMEOUT:
            self.sock.settimeout(self.timeout)

    def __call__(self, *args, **kwargs):
        http.client.HTTPConnection.__init__(self, *args, **kwargs)
        return self


class UnixHTTPHandler(urllib.request.HTTPHandler):
    '''Handler for Unix urls'''

    def __init__(self, unix_socket, **kwargs):
        urllib.request.HTTPHandler.__init__(self, **kwargs)
        self._unix_socket = unix_socket

    def http_open(self, req):
        return self.do_open(UnixHTTPConnection(self._unix_socket), req)



# -*- coding: utf-8 -*-
# Copyright 2020 Matt Martz

from .session import Session


def get(url, **kwargs):
    r"""Sends a GET request. Returns :class:`HTTPResponse` object.

    :arg url: URL to request
    :kwarg \*\*kwargs: Optional arguments that ``Session.request`` takes.
    :returns: requisitor.response.Response
    """

    return Session().get(url, **kwargs)


def options(url, **kwargs):
    r"""Sends a OPTIONS request. Returns :class:`HTTPResponse` object.

    :arg url: URL to request
    :kwarg \*\*kwargs: Optional arguments that ``Session.request`` takes.
    :returns: requisitor.response.Response
    """

    return Session().options(url, **kwargs)


def head(url, **kwargs):
    r"""Sends a HEAD request. Returns :class:`HTTPResponse` object.

    :arg url: URL to request
    :kwarg \*\*kwargs: Optional arguments that ``Session.request`` takes.
    :returns: requisitor.response.Response
    """

    return Session().head(url, **kwargs)


def post(url, data=None, **kwargs):
    r"""Sends a POST request. Returns :class:`HTTPResponse` object.

    :arg url: URL to request.
    :kwarg data: (optional) bytes, or file-like object to send in the body
        of the request.
    :kwarg \*\*kwargs: Optional arguments that ``Session.request`` takes.
    :returns: requisitor.response.Response
    """

    return Session().post(url, data=data, **kwargs)


def put(url, data=None, **kwargs):
    r"""Sends a PUT request. Returns :class:`HTTPResponse` object.

    :arg url: URL to request.
    :kwarg data: (optional) bytes, or file-like object to send in the body
        of the request.
    :kwarg \*\*kwargs: Optional arguments that ``Session.request`` takes.
    :returns: requisitor.response.Response
    """

    return Session().put(url, data=data, **kwargs)


def patch(url, data=None, **kwargs):
    r"""Sends a PATCH request. Returns :class:`HTTPResponse` object.

    :arg url: URL to request.
    :kwarg data: (optional) bytes, or file-like object to send in the body
        of the request.
    :kwarg \*\*kwargs: Optional arguments that ``Session.request`` takes.
    :returns: requisitor.response.Response
    """

    return Session().patch(url, data=data, **kwargs)


def delete(url, **kwargs):
    r"""Sends a DELETE request. Returns :class:`HTTPResponse` object.

    :arg url: URL to request
    :kwargs \*\*kwargs: Optional arguments that ``Session.request`` takes.
    :returns: requisitor.response.Response
    """

    return Session().delete(url, **kwargs)



# -*- coding: utf-8 -*-
# Copyright 2020 Matt Martz

import io
import json
from functools import partial

from .headers import normalize_headers

try:
    import gzip
except ImportError:
    gzip = None


class GzipDecodedResponse(gzip.GzipFile if gzip else object):
    """A file-like object to decode a response encoded with the gzip
        method, as described in RFC 1952.

        Largely copied from ``xmlrpclib``/``xmlrpc.client``
        """
    def __init__(self, response):
        if not gzip:
            raise NotImplementedError

        # response doesn't support tell() and read(), required by
        # GzipFile
        self.io = io.BytesIO()
        for block in iter(partial(response.read, 1024), b''):
            self.io.write(block)
        self.io.seek(0)
        gzip.GzipFile.__init__(self, mode='rb', fileobj=self.io)

    def close(self):
        try:
            gzip.GzipFile.close(self)
        finally:
            self.io.close()


class Response:
    _attr_map = {
        'status_code': 'status',
    }

    def __init__(self, response, request=None, _error=False):

        self._response = response
        if _error:
            self.response = self._response
            self.headers = normalize_headers(self.headers)
        else:
            response.headers = normalize_headers(response.headers)
        self.request = request

        self._set_stream(response)

        self._bytes = None
        self._encoding = None

    def __getattr__(self, name):
        if name == 'read':
            raise AttributeError

        return getattr(
            self._response,
            Response._attr_map.get(name, name)
        )

    def _set_stream(self, stream):
        if self.headers.get('content-encoding', '') == 'gzip':
            self.raw = stream.fp = GzipDecodedResponse(stream)
        else:
            self.raw = stream

    @property
    def encoding(self):
        if self._encoding:
            return self._encoding

        return self.headers.get_param('charset')

    @encoding.setter
    def encoding(self, value):
        self._encoding = value

    @property
    def bytes(self):
        if self._bytes:
            return self._bytes

        self._bytes = self.raw.read()
        return self._bytes

    @property
    def text(self):
        encoding = self.encoding or 'utf-8'
        return self.bytes.decode(encoding)

    def json(self):
        return json.loads(self.text)



# -*- coding: utf-8 -*-
# Copyright 2020 Matt Martz

import mimetypes
from collections import namedtuple
from collections.abc import Mapping
from email.mime.application import MIMEApplication
from email.mime.multipart import MIMEMultipart
from email.mime.nonmultipart import MIMENonMultipart
from email.parser import BytesHeaderParser
from email.policy import HTTP

from .utils import ensure_bytes
from .utils import get_filename
from .utils import is_binary_data
from .utils import read_bytes


Field = namedtuple('Field', ('file', 'content', 'main_type', 'sub_type'))


def _parse_field(field):
    if isinstance(field, Field):
        return field
    elif isinstance(field, str):
        return Field(None, field, 'text', 'plain')
    elif isinstance(field, Mapping):
        file = field.get('file')
        content = field.get('content')
        if not any((file, content)):
            raise ValueError(
                'at least one of "file" or "content" must be provided'
            )

        mime = field.get('mime_type')
        if not mime:
            if not file or isinstance(file, str):
                url = file
            else:
                url = get_filename(file)
            try:
                mime = mimetypes.guess_type(
                    url or '',
                    strict=False
                )[0] or 'application/octet-stream'
            except Exception:
                mime = 'application/octet-stream'
        main_type, _, sub_type = mime.partition('/')
        return Field(file, content, main_type, sub_type)
    else:
        raise TypeError(
            'value must be Field, str, or mapping, cannot be type %s' % (
                field.__class__.__name__
            )
        )


def prepare_multipart(fields):
    """Takes a mapping, and prepares a multipart/form-data body

    :arg fields: Mapping
    :returns: tuple of (content_type, body) where ``content_type`` is
        the ``multipart/form-data`` ``Content-Type`` header including
        ``boundary`` and ``body`` is the prepared bytestring body

    Payload content from a file will be base64 encoded and will include
    the appropriate ``Content-Transfer-Encoding`` and ``Content-Type``
    headers.

    Example:
        {
            "file1": {
                "file": "/bin/true",
                "mime_type": "application/octet-stream"
            },
            "file2": {
                "content": "text based file content",
                "file": "fake.txt",
                "mime_type": "text/plain",
            },
            "text_form_field": "value"
        }
    """

    if not isinstance(fields, Mapping):
        raise TypeError(
            'Mapping is required, cannot be type %s' % (
                fields.__class__.__name__
            )
        )

    m = MIMEMultipart('form-data')
    for field_name in sorted(fields):
        field = _parse_field(fields[field_name])

        if not field.content and field.file:
            content = read_bytes(field.file)
        else:
            content = ensure_bytes(field.content)

        if is_binary_data(content[:1024]):
            part = MIMEApplication(content)
            del part['Content-Type']
            part.add_header(
                'Content-Type',
                '%s/%s' % (field.main_type, field.sub_type)
            )
        else:
            part = MIMENonMultipart(field.main_type, field.sub_type)
            part.set_payload(content)

        part.add_header('Content-Disposition', 'form-data')
        del part['MIME-Version']
        part.set_param(
            'name',
            field_name,
            header='Content-Disposition'
        )
        if field.file:
            part.set_param(
                'filename',
                get_filename(field.file),
                header='Content-Disposition'
            )

        m.attach(part)

    data = m.as_bytes(policy=HTTP)
    del m

    headers, _, content = data.partition(b'\r\n\r\n')
    del data

    parser = BytesHeaderParser().parsebytes
    return (
        parser(headers)['content-type'],
        content
    )



# -*- coding: utf-8 -*-
# Copyright 2020 Matt Martz

import io
import os
import pathlib
import urllib.parse

TEXTCHARS = bytearray(
    {7, 8, 9, 10, 12, 13, 27} | set(range(0x20, 0x100)) - {0x7f}
)


def update_url_params(url, params):
    if isinstance(url, urllib.parse.ParseResult):
        o = url
    else:
        o = urllib.parse.urlparse(url)
    _params = {
        **urllib.parse.parse_qs(o.query),
        **{k: v for k, v in params.items() if v is not None}
    }
    new = o._replace(
        query=urllib.parse.urlencode(
            _params,
            doseq=True
        )
    )
    return new.geturl()


def is_binary_data(data):
    if not isinstance(data, bytes):
        raise TypeError(
            'value must be a bytes, cannot be type %s' % (
                data.__class__.__name__
            )
        )
    return bool(
        data.translate(None, TEXTCHARS)
    )


def is_binary_fileobj(obj):
    return isinstance(obj, (io.RawIOBase, io.BufferedIOBase))


def get_filename(obj, basename=True):
    if isinstance(obj, pathlib.Path):
        return obj.name if basename else str(obj)
    elif is_binary_fileobj(obj):
        return os.path.basename(obj.name) if basename else obj.name
    elif isinstance(obj, str):
        return os.path.basename(obj) if basename else obj
    else:
        raise TypeError(
            'value must be a str, pathlib.Path, or binary file object, '
            'cannot be type %s' % (
                obj.__class__.__name__
            )
        )


def read_bytes(obj):
    if is_binary_fileobj(obj):
        return obj.read()
    elif isinstance(obj, pathlib.Path):
        return obj.read_bytes()
    elif isinstance(obj, str):
        return pathlib.Path(obj).read_bytes()
    else:
        raise TypeError(
            'value must be a str, pathlib.Path, or binary file object, '
            'cannot be type %s' % (
                obj.__class__.__name__
            )
        )


def ensure_bytes(obj, encoding='utf-8'):
    if isinstance(obj, bytes):
        return obj
    return obj.encode(encoding)



# -*- coding: utf-8 -*-
# Copyright 2020 Matt Martz

import http.client

from .sentinel import Sentinel


class Headers(http.client.HTTPMessage):
    def __repr__(self):
        return repr(dict(self))

    def copy(self):
        headers = Headers()
        headers.update(self)
        return headers

    def pop(self, key, default=Sentinel):
        value = self.get(key, default)
        if value is Sentinel:
            raise KeyError(key)
        del self[key]
        return value

    def update(self, *args, **kwds):
        if len(args) > 1:
            raise TypeError('update expected at most 1 arguments, got %d' %
                            len(args))

        items = []
        if args:
            other = args[0]
            if isinstance(other, (dict, http.client.HTTPMessage)):
                items[:] = other.items()
            elif isinstance(other, list):
                items[:] = other
            else:
                raise TypeError
        for obj in (items, kwds.items()):
            for name, value in obj:
                self.pop(name, None)
                self[name] = value

    def add_header(self, name, value, **params):
        orig = self.pop(name, None)
        super(Headers, self).add_header(name, value, **params)
        if orig:
            new = self.pop(name)
            self[name] = '{}, {}'.format(orig, new)


def normalize_headers(headers):
    normalized = Headers()

    if isinstance(headers, (dict, http.client.HTTPMessage)):
        items = headers.items()
    elif isinstance(headers, list):
        items = headers
    else:
        raise TypeError

    # Don't be lossy, append header values for duplicate headers
    for name, value in items:
        if name in normalized:
            old = normalized[name]
            del normalized[name]
            normalized[name] = ', '.join((old, value))
        else:
            normalized[name] = value

    return normalized



# -*- coding: utf-8 -*-
# Copyright 2020 Matt Martz

import http.client
import http.cookiejar
import json as _json
import ssl
import urllib.request
from urllib.parse import urlparse

from .auth import validate_auth
from .handlers import HTTPErrorHandler
from .handlers import HTTPSClientAuthHandler
from .handlers import RedirectHandler
from .handlers import UnixHTTPHandler
from .headers import Headers
from .headers import normalize_headers
from .multipart import prepare_multipart
from .response import Response
from .sentinel import Sentinel
from .utils import update_url_params


def _validate_cert(value):
    if value in (None, Sentinel):
        return value
    elif isinstance(value, tuple) and len(value) == 2:
        return value
    raise TypeError


class Session:
    def __init__(self):
        self._auth = None
        self._cert = None
        self._headers = Headers()

        self.cookies = http.cookiejar.CookieJar()

        self.handlers = []
        self.params = {}
        self.unix_socket = None
        self.verify = True

    @property
    def auth(self):
        return self._auth

    @auth.setter
    def auth(self, value):
        self._auth = validate_auth(value)

    @property
    def cert(self):
        return self._cert

    @cert.setter
    def cert(self, value):
        self._cert = _validate_cert(value)

    @property
    def headers(self):
        return self._headers

    @headers.setter
    def headers(self, value):
        self._headers = normalize_headers(value)

    def _fallback(self, value, fallback):
        if value is Sentinel:
            return fallback
        return value

    def _create_context(self, verify):
        context = ssl.create_default_context()
        if not verify:
            context.check_hostname = False
            context.verify_mode = ssl.CERT_NONE

        if verify and verify is not True:
            context.load_verify_locations(cafile=verify)

        return context

    def request(self, method, url, params=Sentinel, data=None,
                headers=Sentinel, cookies=Sentinel, files=None, auth=Sentinel,
                timeout=None, allow_redirects=True, verify=Sentinel,
                cert=Sentinel, json=Sentinel, unix_socket=Sentinel):

        if sum(bool(x) for x in (data, json, files)) > 1:
            raise TypeError(
                '"data", "json", and "files" are mutually exclusive'
            )

        _headers = self.headers.copy()
        _headers.update(headers or {})

        _params = self.params.copy()
        _params.update(params or {})

        auth = self._fallback(auth, self.auth)
        cert = _validate_cert(self._fallback(cert, self.cert)) or (None, None)
        cookies = self._fallback(cookies, self.cookies)
        unix_socket = self._fallback(unix_socket, self.unix_socket)
        verify = self._fallback(verify, self.verify)

        if json is not Sentinel:
            if 'content-type' not in _headers:
                _headers['content-type'] = 'application/json'
            encoding = _headers.get_param('charset', 'utf-8')
            data = _json.dumps(json).encode(encoding)

        if files:
            content_type, data = prepare_multipart(files)
            _headers['content-type'] = content_type

        https_kwargs = {
            'context': self._create_context(verify),
            'client_cert': cert[0],
            'client_key': cert[1],
            'unix_socket': unix_socket,
        }

        handlers = [
            HTTPSClientAuthHandler(**https_kwargs),
            RedirectHandler(allow_redirects=allow_redirects),
            HTTPErrorHandler,
            urllib.request.HTTPCookieProcessor(cookies),
        ]

        if unix_socket:
            handlers.append(UnixHTTPHandler(unix_socket))

        o = urlparse(url)
        if auth:
            auth_info = validate_auth(auth)(o)
            _headers.update(
                auth_info.get('headers', {})
            )
            handlers.extend(
                auth_info.get('handlers', [])
            )

        handlers.extend(self.handlers)

        opener = urllib.request.build_opener(*handlers)
        req = urllib.request.Request(
            update_url_params(o, _params),
            data=data or None,
            method=method,
            headers=_headers,
        )
        return Response(
            opener.open(
                req,
                timeout=timeout,
            ),
            request=req
        )

    def get(self, url, **kwargs):
        r"""Sends a GET request. Returns :class:`HTTPResponse` object.

        :arg url: URL to request
        :kwarg \*\*kwargs: Optional arguments that ``open`` takes.
        :returns: requisitor.response.Response
        """

        return self.request('GET', url, **kwargs)

    def options(self, url, **kwargs):
        r"""Sends a OPTIONS request. Returns :class:`HTTPResponse` object.

        :arg url: URL to request
        :kwarg \*\*kwargs: Optional arguments that ``open`` takes.
        :returns: requisitor.response.Response
        """

        return self.request('OPTIONS', url, **kwargs)

    def head(self, url, **kwargs):
        r"""Sends a HEAD request. Returns :class:`HTTPResponse` object.

        :arg url: URL to request
        :kwarg \*\*kwargs: Optional arguments that ``open`` takes.
        :returns: requisitor.response.Response
        """

        return self.request('HEAD', url, **kwargs)

    def post(self, url, data=None, **kwargs):
        r"""Sends a POST request. Returns :class:`HTTPResponse` object.

        :arg url: URL to request.
        :kwarg data: (optional) bytes, or file-like object to send in the body
            of the request.
        :kwarg \*\*kwargs: Optional arguments that ``open`` takes.
        :returns: requisitor.response.Response
        """

        return self.request('POST', url, data=data, **kwargs)

    def put(self, url, data=None, **kwargs):
        r"""Sends a PUT request. Returns :class:`HTTPResponse` object.

        :arg url: URL to request.
        :kwarg data: (optional) bytes, or file-like object to send in the body
            of the request.
        :kwarg \*\*kwargs: Optional arguments that ``open`` takes.
        :returns: requisitor.response.Response
        """

        return self.request('PUT', url, data=data, **kwargs)

    def patch(self, url, data=None, **kwargs):
        r"""Sends a PATCH request. Returns :class:`HTTPResponse` object.

        :arg url: URL to request.
        :kwarg data: (optional) bytes, or file-like object to send in the body
            of the request.
        :kwarg \*\*kwargs: Optional arguments that ``open`` takes.
        :returns: requisitor.response.Response
        """

        return self.request('PATCH', url, data=data, **kwargs)

    def delete(self, url, **kwargs):
        r"""Sends a DELETE request. Returns :class:`HTTPResponse` object.

        :arg url: URL to request
        :kwargs \*\*kwargs: Optional arguments that ``open`` takes.
        :returns: requisitor.response.Response
        """

        return self.request('DELETE', url, **kwargs)



# -*- coding: utf-8 -*-
# Copyright 2020 Matt Martz

import base64
import urllib.request


def validate_auth(value):
    if value is None:
        return value
    elif isinstance(value, tuple):
        if len(value) != 2:
            raise TypeError
        return HTTPBasicAuth(*value)
    elif isinstance(value, HTTPAuth):
        return value

    raise TypeError


class HTTPAuth:
    def __init__(self, user, password):
        self.user = user
        self.password = password


class HTTPBasicAuth(HTTPAuth):
    def __call__(self, parsed_url):
        return {
            'headers': {
                'Authorization': 'Basic {}'.format(
                    base64.b64encode(
                        '{}:{}'.format(
                            self.user,
                            self.password
                        ).encode()
                    ).decode()
                ),
            }
        }


class HTTPDigestAuth(HTTPAuth):
    def __call__(self, parsed_url):
        pwd_mgr = urllib.request.HTTPPasswordMgrWithDefaultRealm()
        pwd_mgr.add_password(None, parsed_url.netloc, self.user, self.password)

        return {
            'handlers': [
                urllib.request.HTTPDigestAuthHandler(pwd_mgr),
            ]
        }



# -*- coding: utf-8 -*-
# Copyright 2020 Matt Martz

import urllib.error

from .response import Response


class HTTPError(urllib.error.HTTPError, Response):
    def __init__(self, url, code, msg, hdrs, fp, req=None):
        super().__init__(url, code, msg, hdrs, fp)
        Response.__init__(self, fp, req, _error=True)
